#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2022-02-15
#
# WATO file for curl plugin (bakery and check)
#
# 2022-02-19: moved global options under "Default settings"
#             added per url settings
#             added proxy settings (--proxy, --proxy-user, --proxy-digest, --proxy-basic, --proxy-ntlm, --proxy-anyauth)
#             added follow redirects (--location)
#             added don't verify certificates (--insecure)
# 2022-02-21: added tls/ssl version (--tlsv1.3, --tlsv1.2, --tlsv1.1, --tlsv1.0, --tlsv1,  --sslv3, --sslv2)
#             added http version (--http2, --http1.1, --http1.0)
#             added header only (--head)
#             added user authentication (--user, --basic, --digest, --ntlm, --ntlm-wb, --negotiate, --anyauth)
#             added to proxy authentication (--proxy-negotiate)
# 2022-02-22: added option for cURL check plugin
# 2022-02-24: changed forbidden_chars from '<>| ' to '"<>#%{}|\^~[]` ' + "'"
# 2022-02-25: added plugin interval and timeout options
#             added noproxy option to default settings
# 2022-02-26: added proxy protocol (--socks4, --socks4a, --socks4a, --socks5-hostname)
# 2022-02-27: added expected_strings option to default settings
# 2022-02-28: added expected_strings option to url settings
#             added state_item_not_found option
# 2022-03-01: added options --limit-rate, --max-filesize, --max-time, --speed-limit, --speed-time, --connect-timeout
#             added options --user-agent, --compressed, --no-alpn, --no-npn, --tcp-fastopen, --tcp-nodelay,
#             added options --cert-status
# 2022-03-02: added options --referer -header and api_key_header (header from password store)
# 2022-03-05: added option --dump-header, state_header_str_not_found
# 2022-03-06: added option --verbose, --stderr cert_time_left
#             fixed upper/lower levels
# 2022-03-10: optimized generation of upper/lower limit
#             added transform for ms to s -> no need in th agent to change the value any more
# 2022-03-11: added check to avoid duplicate service names in url list
#             added --location-trust, --max-redirs
#             reworked redirect (--location, --location-trust, --max-redirs)
#             reworked cert verify (-insecure, --cert-status, --ssl-no-revok)
#             reworked advanced_settings (--no-alpn, --no-npn, --tcp-fastopen, -tcp-nodelay)
#             removed get_session_data. moved to curl default options
# 2022-03-12: added --cacert
#             added max_age
# 2022-03-13: added post data
#             made some entries fordable (Default options/Per URL settings, CA Cert)
#             changed url and service_name to curl_service tuple
#             changed headers to list of tuple
# 2022-03-15: added regex pattern match for bakery
# 2022-03-18: added regex pattern match for check
# 2022-03-19: added options --dns-interface, -dns-ipv4-addr, --dns-ipv6-addr and --dns-servers
#             reworked ip_address_resolution section
#             added options --ftp-account, --ftp-alternative-to-user, --ftp-create-dirs, --ftp-method
#             added options --ftp-pasv, --disable-epsv, --ftp-pret, --ftp-skip-pasv-ip
#             added options --ftp-port, --disable-eprt
#             added options --ftp-ssl-control, --ftp-ssl-ccc, --ftp-ssl-ccc-mode
# 2022-03-21: moved  --connect-timeout, --limit-rate, --max-filesize, --max-time, --speed-limit, --speed-time
#             to "limits" sub Directory
# 2022-03-22: added curl_error_code_to_ignore and http_error_code_to_ignore options
# 2022-03-24: added options --hostpubmd5, --hostpubsha256, --pubkey
# 2022-03-24: added options --key --passs
#             reworked user_auth section
# 2022-03-25: added options --compressed-ssh, --list-only, --use-ascii
#             added options --path-as-is, --ssl-allow-beast, --no-buffer, --no-keepalive, --no-sessionid
# 2022-03-28: added options --mail-auth, --mail-from, --mail-rcpt, --mail-rcpt-allowfails, --crlf, --upload-file (SMTP)
# 2022-04-10: added options to deploy cURL executable, no separate rules for curl and curl executable anymore
#             windows/linux summarized under on option, no separate WATO rules per OS necessary anymore
# 2022-04-26: added check option clickable_url
#             clarified option 'Don\'t verify certificate/pub key', 'Don\'t stop at verify errors (certificate/pub key)'
# 2023-10-21: split in service and bakery part
#             modified for cmk 2.2.0

from cmk.gui.i18n import _
from cmk.gui.valuespec import (
    Dictionary,
    TextUnicode,
    FixedValue,
    Integer,
    Tuple,
    MonitoringState,
    ListOfStrings,
    Checkbox,
    Age,
)
from cmk.gui.plugins.wato.utils import (
    rulespec_registry,
    RulespecGroupCheckParametersOperatingSystem,
    CheckParameterRulespecWithItem,
)

bakery_plugin_version = '20220426.v0.0.8'

# unsafe characters https://www.tutorialspoint.com/html/html_url_encoding.htm
forbidden_chars = '"<>#%{}|\^~[]` \''

_limits = [
    ('time_namelookup', 'Time name lookup', 'ms', ),
    ('time_connect', 'Time connect', 'ms', ),
    ('time_appconnect', 'Time app connect', 'ms', ),
    ('time_pretransfer', 'Time pre transfer', 'ms', ),
    ('time_redirect', 'Time redirect', 'ms', ),
    ('time_starttransfer', 'Time start transfer', 'ms', ),
    ('time_total', 'Time Total', 'ms', ),
    ('cert_time_left', 'Certificate life time', 'Day(s)',),
    ('num_connects', '# of connects', 'count',),
    ('num_redirects', '# of redirects', 'count',),
    ('num_headers', '# of headers', 'count',),
    ('size_download', 'Size download', 'bytes',),
    ('size_upload', 'Size upload', 'bytes',),
    ('size_header', 'Size header', 'bytes',),
    ('size_request', 'Size request', 'bytes',),
    ('speed_download', 'Speed download', 'bytes/s',),
    ('speed_upload', 'Speed upload', 'bytes/s',),
]

_curl_check_elements = [
    ('clickable_url',
     FixedValue(
         True,
         title=_('Render clickable URLs for HTTP/HTTPS (see Inline help)'),
         totext=_(''),
         help=_('Needs "Escape HTML in service output" rule or in Global settings enabled to work.')
     )),
    ('state_item_not_found',
     MonitoringState(
         default_value=1,
         title=_('State on item not found'),
         help=_('Monitoring state if the item is not found in the agent data. Default is UNKNOWN.')
     )),
    ('state_http_result_not_200',
     MonitoringState(
         default_value=1,
         title=_('State on HTTP result not OK'),
         help=_('Monitoring state if the HTTP return code is not in 2xx or 3xx. Default is WARN.')
     )),
    ('http_error_code_to_ignore',
     ListOfStrings(
         title=_('HTTP error codes to ignore'),
         allow_empty=False,
         orientation='horizontal',
         # valuespec=Integer(size=3, minvalue=0, maxvalue=999),
     )),
    ('state_curl_result_not_0',
     MonitoringState(
         default_value=1,
         title=_('State on cURL exit code not OK'),
         help=_('Monitoring state if the exit code is not 0. Default is WARN.')
     )),
    ('curl_error_code_to_ignore',
     ListOfStrings(
         title=_('cURL error codes to ignore'),
         allow_empty=False,
         orientation='horizontal',
         # valuespec=Integer(size=3, minvalue=0, maxvalue=999),
     )),
    ('state_verify_sll_not_0',
     MonitoringState(
         default_value=1,
         title=_('State on SSL verify not OK'),
         help=_('Monitoring state if the SSL verify code is not 0. Default is WARN.')
     )),
    ('state_expected_str_not_found',
     MonitoringState(
         default_value=1,
         title=_('State on expected string not found'),
         help=_('Monitoring state if one expected string is not found in the cURL output. Default is WARN.')
     )),
    ('state_header_str_not_found',
     MonitoringState(
         default_value=1,
         title=_('State on expected header not found'),
         help=_('Monitoring state if one expected header string is not found in the cURL output. Default is WARN.')
     )),
    ('state_for_regex',
     Tuple(
         title=_('State for regex pattern match'),
         elements=[
             MonitoringState(
                 default_value=0,
                 title=_('State on regex pattern match'),
             ),
             MonitoringState(
                 default_value=1,
                 title=_('State on regex pattern don\'t match'),
             ),
             MonitoringState(
                 default_value=0,
                 title=_('State on regex pattern match info missing'),
             ),
         ],
     )),
    ('show_additional_info',
     Tuple(
         title=_('Show additional info'),
         help=_('Shows RAW data from cURL in the service details.'),
         elements=[
             Checkbox('Request headers'),
             Checkbox('Response headers'),
             Checkbox('(TLS/SSL) session info'),
             Checkbox('RAW data'),
         ],
     )),
    ('max_age',
     Tuple(
         title=_('Maximum age'),
         elements=[
             Age(
                 title=_('Warning at'),
                 help=_('Warn, if the age of the page is older than this'),
                 default_value=3600 * 24,
             ),
             Age(
                 title=_('Critical at'),
                 help=_('Critical, if the age of the page is older than this'),
                 default_value=3600 * 24 * 4,
             ),
             MonitoringState(
                 default_value=0,
                 title=_('State if not available'),
                 help=_(
                     'Monitoring state if the "Last-Modified" header is not available in the agent data. Default is OK.'
                 )
             ),
         ]
     )),
]


def _get_tuple_upper(_unit: str) -> Tuple:
    return Tuple(
        title=_('Upper limits'),
        orientation='horizontal',
        elements=[
            Integer(title=_('Warning at'), minvalue=0, unit=_(_unit), size=10),
            Integer(title=_('Critical at'), minvalue=0, unit=_(_unit), size=10),
        ],
    )


def _get_tuple_lower(_unit):
    return Tuple(
        title=_('Lower limits'),
        orientation='horizontal',
        elements=[
            Integer(title=_('Warning below'), minvalue=0, unit=_(_unit), size=10),
            Integer(title=_('Critical below'), minvalue=0, unit=_(_unit), size=10),
        ],
    )


for key, label, unit in _limits:
    _curl_check_elements.append(
        (key,
         Dictionary(
             title=_(label),
             elements=[
                 ('upper', _get_tuple_upper(unit),),
                 ('lower', _get_tuple_lower(unit),),
             ],
         ))
    )


def _valuespec_curl():
    return Dictionary(
        elements=_curl_check_elements,
    )


rulespec_registry.register(
    CheckParameterRulespecWithItem(
        check_group_name='curl',
        group=RulespecGroupCheckParametersOperatingSystem,
        parameter_valuespec=_valuespec_curl,
        title=lambda: _('cURL'),
        match_type='dict',
        item_spec=lambda: TextUnicode(title=_('cURL service name'), ),
    ))
